#include <sys/types.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>

char *pgmname;
extern char *getenv();

#define	APPSOCKETPORT	9100
#define	BUFFERSIZE	16384

int file, sock;
extern cleanup();
extern catcher();

main(argc, argv)
int argc;
char **argv;
{
    char *printername;
    int nflag;
    struct sockaddr_in printer;
    int errs = 0;

    pgmname = argv[0];

    /*
     * Look for a "-Pprintername" argument.  If not found,
     * get the printer name from the environment.  If still
     * not found, print an error message and give up.
     *
     * Also look for a "-n" argument.  If found, don't
     * read data from the back channel.
     */
    nflag = 0;
    printername = NULL;
    while (argc > 1 && argv[1][0] == '-') {
	--argc;
	++argv;
	if (strcmp(*argv, "-n") == 0) {
	    nflag++;
	} else if (strncmp(*argv, "-P", 2) == 0) {
	    (*argv) += 2;
	    if (**argv == '\0') {
		argc--;
		argv++;
	    }
	    printername = *argv;
	} else {
	    fprintf(stderr, "usage: asprint [-n] [-Pprinter] [file ...]\n");
	    return -1;
	}
    }
    if (printername == NULL) {
	printername = getenv("PRINTER");
    }
    if (printername == NULL || *printername == '\0') {
	fprintf(stderr,
	  "%sock: no printer?\n", pgmname);
	fprintf(stderr,
	  "Specify -Pprinter or set the PRINTER environment variable.\n");
	return -1;
    }

    /*
     * Convert printer name to Internet address and port number.
     */
    if (lookup(printername, &printer) < 0) {
	fprintf(stderr, "Can't find address for \"%sock\".\n", printername);
	return -1;
    }

    /*
     * Loop through the remaining arguments, sending each file
     * to the printer in turn.
     */
    if (argc == 1) {
	/* no arguments -- read from stdin */
	file = 0;
	if (client(&printer, nflag) < 0)
	    return -1;
    } else {
	while (--argc > 0) {
	    if (strcmp(*++argv, "-") == 0) {
		file = 0;
		if (client(&printer, nflag) < 0)
		    return -1;
	    } else {
		if ((file = open(*argv, 0)) < 0) {
		    perror(*argv);
		    errs++;
		} else {
		    if (client(&printer, nflag) < 0)
			return -1;
		    close(file);
		}
	    }
	}
    }
    return errs;
}

/*
 * Look up the address for the printer.  We allow either a hostname or
 * an internet address in dot notation.  Either one can be followed by
 * a comma and a TCP port number.
 */
lookup(name, addr)
char *name;
struct sockaddr_in *addr;
{
    struct hostent *hostent;
    int b0, b1, b2, b3;
    char *ptr;

    addr->sin_family = AF_INET;

    /*
     * First, look for a trailing comma and port number.
     * If found, change the comma to a NUL and convert the
     * port number to machine format.
     */
    if (ptr = strrchr(name, ',')) {
	*ptr++ = '\0';
	addr->sin_port = atoi(ptr);
    } else {
	addr->sin_port = htons(APPSOCKETPORT);
    }

    /*
     * If the name looks like an Internet address, pack the
     * four byte values into a single machine word.
     */
    if (sscanf(name, "%d.%d.%d.%d", &b0, &b1, &b2, &b3) == 4) {
	/* Pack four bytes into machine word in network order.  Doing it
	   this way makes us independent of machine-to-machine differences
	   in the declaration of struct in_addr. */
	((char *)&addr->sin_addr)[0] = b0;
	((char *)&addr->sin_addr)[1] = b1;
	((char *)&addr->sin_addr)[2] = b2;
	((char *)&addr->sin_addr)[3] = b3;
	return 0;
    }

    /*
     * Try looking up the name the usual way.
     */
    if ((hostent = gethostbyname(name)) != NULL &&
      hostent->h_addrtype == AF_INET &&
      hostent->h_length == sizeof(struct in_addr)) {
	addr->sin_addr = *(struct in_addr *)hostent->h_addr;
	return 0;
    }

    return -1;
}

client(printer, nflag)
struct sockaddr_in *printer;
int nflag;
{
    int limit;
    struct linger linger;
    int zero = 0, one = 1;
    struct channel {
	int fdread, fdwrite;
	int eofseen, eofsent;
	char buf[BUFFERSIZE], *ptr;
	int count;
    } fwd, back;
    fd_set readfds, writefds, exceptfds;
    int n;

    /*
     * Open a connection to the printer.
     */
    if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("socket");
	return -1;
    }
    if (connect(sock, printer, sizeof *printer) < 0) {
	perror("connect");
	return -1;
    }

    /*
     * Ask for large send and receive queues.
     */
    limit = BUFFERSIZE;
    if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &limit, sizeof limit) < 0) {
	perror("sndbuf");
	return -1;
    }
    if (!nflag) {
	limit = BUFFERSIZE;
	if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &limit, sizeof limit) < 0) {
	    perror("rcvbuf");
	    return -1;
	}
    }
    
    /*
     * Arrange to send abrupt close (RST) if we're interrupted.
     */
    linger.l_onoff = 1;
    linger.l_linger = 0;
    if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger) < 0) {
	perror("linger");
	return -1;
    }

    /*
     * Set up traps to disable non-blocking I/O if we're interrupted.
     * This helps to prevent problems when we return to the shell.
     */
    on_exit(cleanup, NULL);
    signal(SIGHUP, catcher);
    signal(SIGINT, catcher);
    signal(SIGQUIT, catcher);
    signal(SIGTERM, catcher);

    /*
     * Enable non-blocking I/O on all three files.
     */
    setnbio(1);

    /*
     * Initialize state for the forward and back channels.
     */
    fwd.fdread = file;
    fwd.fdwrite = sock;
    fwd.eofseen = fwd.eofsent = 0;
    fwd.ptr = fwd.buf;
    fwd.count = 0;

    back.fdread = sock;
    back.fdwrite = 1;
    back.eofseen = back.eofsent = 0;
    back.ptr = back.buf;
    back.count = 0;

    if (nflag) {
	back.eofseen = back.eofsent = 1;
    }
    /*
     * Clear all the descriptor masks.
     *
     * We set bits to select for, then clear them as we go through
     * the body of the loop, so that all the masks are clear again
     * at the top of the loop.
     *
     * We wouldn't need an exception mask except that some
     * (broken) implementations don't check for a NULL pointer.
     */
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&exceptfds);

    /*
     * This is the main data mover loop.  Keep running until
     * we've seen and delivered an EOF in both directions.
     */
    while (!fwd.eofsent || !back.eofsent) {
	/*
	 * If the forward channel buffer is empty, try to fill it.
	 * If not, try to empty it.
	 */
	if (!fwd.eofsent) {
	    if (fwd.count == 0 && !fwd.eofseen)
		FD_SET(fwd.fdread, &readfds);
	    else
		FD_SET(fwd.fdwrite, &writefds);
	}
	/*
	 * If the back channel buffer is empty, try to fill it.
	 * If not, try to empty it.
	 */
	if (!back.eofsent) {
	    if (back.count == 0 && !back.eofseen)
		FD_SET(back.fdread, &readfds);
	    else
		FD_SET(back.fdwrite, &writefds);
	}
	/*
	 * Wait here until one of the read descriptors has data
	 * available, or one of the write descriptors has buffer
	 * space available.
	 */
	if (select(FD_SETSIZE, &readfds, &writefds, &exceptfds, NULL) <= 0) {
	    perror("select");
	    return -1;
	}
	/*
	 * If we were waiting for forward channel data and it's now
	 * available, fill the buffer.
	 */
	if (FD_ISSET(fwd.fdread, &readfds)) {
	    FD_CLR(fwd.fdread, &readfds);
	    fwd.ptr = fwd.buf;
	    if ((fwd.count = read(fwd.fdread, fwd.buf, BUFFERSIZE)) < 0) {
		perror("fwd read");
		return -1;
	    }
	    if (fwd.count == 0) {
		fwd.eofseen = 1;
	    }
	}
	/*
	 * If we were waiting for the forward channel to drain and
	 * there's more space available, empty the buffer.
	 */
	if (FD_ISSET(fwd.fdwrite, &writefds)) {
	    FD_CLR(fwd.fdwrite, &writefds);
	    if (fwd.count > 0) {
		if ((n = write(fwd.fdwrite, fwd.ptr, fwd.count)) < 0) {
		    perror("fwd write");
		    return -1;
		}
		fwd.ptr += n;
		fwd.count -= n;
	    } else {
		/* We couldn't have gotten here unless there was
		   something to be written.  Since the buffer's
		   empty, it must have been an EOF. */
		if (!fwd.eofseen) {
		    fprintf(stderr, "fwd no data and no EOF (can't happen)\n");
		    return -1;
		}
		if (shutdown(fwd.fdwrite, 1) < 0) {
		    perror("fwd shutdown");
		    return -1;
		}
#ifdef DEBUG
		fprintf(stderr, "transmission complete; waiting for printer\n");
#endif
		fwd.eofsent = 1;
	    }
	}
	/*
	 * If we were waiting for back channel data and it's now
	 * available, fill the buffer.
	 */
	if (FD_ISSET(back.fdread, &readfds)) {
	    FD_CLR(back.fdread, &readfds);
	    back.ptr = back.buf;
	    if ((back.count = read(back.fdread, back.buf, BUFFERSIZE)) < 0) {
		perror("back read");
		return -1;
	    }
	    if (back.count == 0) {
		back.eofseen = 1;
	    }
	}
	/*
	 * If we were waiting for the back channel to drain and
	 * there's more space available, empty the buffer.
	 */
	if (FD_ISSET(back.fdwrite, &writefds)) {
	    FD_CLR(back.fdwrite, &writefds);
	    if (back.count > 0) {
		if ((n = write(back.fdwrite, back.ptr, back.count)) < 0) {
		    perror("back write");
		    return -1;
		}
		back.ptr += n;
		back.count -= n;
	    } else {
		/* We couldn't have gotten here unless there was
		   something to be written.  Since the buffer's
		   empty, it must have been an EOF. */
		if (!back.eofseen) {
		    fprintf(stderr, "back no data and no EOF (can't happen)\n");
		    return -1;
		}
		back.eofsent = 1;
	    }
	}
    }

    /*
     * Normal completion.  We've already exchanged FINs (and probably
     * last ACK), so we expect the close to complete immediately whether
     * or not linger is set.  We still need to turn off the option in
     * order to generating a spurious reset after the last ACK.
     */
    setnbio(0);

    linger.l_onoff = 0;
    linger.l_linger = 0;
    if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger) < 0) {
	perror("linger");
	return -1;
    }

    if (close(sock) < 0) {
	perror("close");
	return -1;
    }

    return 0;
}

setnbio(flag)
int flag;
{
    (void) ioctl(file, FIONBIO, &flag);
    (void) ioctl(sock, FIONBIO, &flag);
    (void) ioctl(1, FIONBIO, &flag);
}

cleanup()
{
    setnbio(0);
}

catcher()
{
    cleanup();
    exit(1);
}
