#include <sys/types.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netdb.h>

#define	HUGE	65536

main(argc, argv)
int argc;
char **argv;
{
    int s;

    if (argc < 4) {
	fprintf(stderr, "usage: lpdsend server queue file ...\n");
	exit(1);
    }

    if ((s = openserver(argv[1], argv[2])) < 0)
	exit(2);

    argc -= 2;
    argv += 2;

    while (--argc) {
	if (sendfile(s, *++argv) < 0)
	    exit(3);
    }

    close(s);
    exit(0);
}

openserver(servername, queuename)
char *servername, *queuename;
{
    struct hostent *server;
    struct servent *service;
    struct sockaddr_in sin;
    int s;
    char ack;

    if ((server = gethostbyname(servername)) == NULL) {
	perror(servername);
	return -1;
    }

    if ((service = getservbyname("printer", "tcp")) == NULL) {
	perror("lpd");
	return -1;
    }

    if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("socket");
	return -1;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = service->s_port;
    sin.sin_addr.s_addr = ((struct in_addr *)server->h_addr_list[0])->s_addr;

    if (connect(s, &sin, sizeof sin) < 0) {
	perror("connect");
	return -1;
    }

    if (write(s, "\02", 1) != 1		/* print job header */
     || write(s, queuename, strlen(queuename)) != strlen(queuename)
     || write(s, "\n", 1) != 1) {
	perror("socket write (job header)");
	return -1;
    }

    if (read(s, &ack, 1) != 1) {
	perror("socket read (job ACK)");
	return -1;
    }

    if (ack != '\0') {
	fprintf(stderr, "%s: queue down; returned %d\n",
		queuename, ack & 0xff);
	return -1;
    }

    return s;
}

sendfile(s, fname)
int s;
char *fname;
{
    int f;
    struct stat st;
    char buf[HUGE];
    int n;
    static int num = 0;

    if (stat(fname, &st) < 0) {
	perror(fname);
	return -1;
    }

    if ((f = open(fname, O_RDONLY, 0)) < 0) {
	perror(fname);
	return -1;
    }

    sprintf(buf, "\003%d dfa%03dlpdsend\n", st.st_size, ++num);
    if (write(s, buf, strlen(buf)) != strlen(buf)) {
	perror("socket write (file header)");
	close(f);
	return -1;
    }

    if (read(s, buf, 1) != 1) {
	perror("socket read (file ACK)");
	close(f);
	return -1;
    }

    if (buf[0] != '\0') {
	fprintf(stderr, "%s: bad file name; returned %d\n",
		fname, buf[0] & 0xff);
	close(f);
	return -1;
    }

    while (n = read(f, buf, HUGE)) {
	if (n < 0) {
	    perror(fname);
	    break;
	}
	if (write(s, buf, n) != n) {
	    perror("socket write (data)");
	    close(f);
	    return -1;
	}
    }

    if (write(s, "\0", 1) != 1) {
	perror("socket write (EOF)");
	close(f);
	return -1;
    }

    if (read(s, buf, 1) != 1) {
	perror("socket read (EOF ACK)");
	close(f);
	return -1;
    }

    if (buf[0] != '\0') {
	fprintf(stderr, "%s: failed; returned %d\n",
		fname, buf[0] & 0xff);
	close(f);
	return -1;
    }

    close(f);
    return 0;
}

